unit CRYPT2;
  {V2.25/11.91-11.99/U.Gschwind.
  Kryptografische Routinen.
  16Bit DOS-Version/Turbo Pascal 6.0.

  MD5. Einweg-Hashfunktion, Berechnung eines Hashwertes von 128 Bit Laenge.
  Dem Programmcode liegt der folgende Algorithmus zugrunde:
    "MD5 Message-Digest, RFC 1321, 1992. Algorithmus von Ron Rivest".
  Beispiele fuer MD5-konforme Hashwerte:
    ''  = d41d8cd98f00b204e9800998ecf8427e
    'a' = 0cc175b9c0f1b6a831c399e269772661.
  Aufruffolge:
    1. md5Init(pMessage,MessageLength)
    2. md5Update(MessageBufSize), -> bei Dateien: wiederholen bis Dateipuffer leer
    3. md5Finals(pHashValue)
  }

interface

const
  Crypt2Ver='2.25';                {TPU-Version}
  md5HashSize=16;                  {MD5: Hashwertlaenge 16 Byte/128 Bit}
  md5BlockSize=64;                 {MD5: MD5-Blockgroesse (Input Buffer) 64 Byte/512 Bit}
  md5MagicNumber:Array[1..md5HashSize div 4] of longint=
    ($67452301,$efcdab89,$98badcfe,$10325476);
                                   {MD5: Initialisierung mit Magic-Zahlen entspr.}
                                   {RSA Data Security, Inc., 1992, RFC 1321.}

procedure md5Init(pMessage:pointer;MessageLength:longint);
procedure md5Update(MessageBufSize:longint);
procedure md5Finals(pHashValue:pointer);
function  md5Valid:boolean;


implementation

type
  md5Rec=record
    A,B,C,D:longint;               {MD5: Hashwert (16 Byte/128 Bit)}
    M:Array[1..md5BlockSize div 4] of longint;
                                   {MD5: Input Buffer (MD5-Block 64 Byte/512 Bit)}
    UpdateLoopNo:longint           {MD5: Update-Durchlaeufe}
  end;
  md5MsgRec=record                 {MD5: Eingangstext-Pufferdaten}
    pStart,pRest:pointer;
    Length,LengthRest:longint
  end;
var
  MD:md5Rec;
  Msg:md5MsgRec;


procedure md5Transform;forward;
procedure md5TestShowMD(pABCD,pMD5Block:pointer;md5BlockSize,Len:longint);forward;


procedure md5Init(pMessage:pointer;MessageLength:longint);
  {pMessage: Zeiger auf Text-Puffer, -> bei Dateien: Zeiger auf Dateipuffer,
  MessageLength: Textlaenge, -> bei Dateien: Dateigroesse}
begin {md5Init()}
  Msg.pStart:=pMessage;            {Eingangstext-Pufferdaten setzen}
  Msg.pRest:=Msg.pStart;
  Msg.Length:=MessageLength;
  Msg.LengthRest:=Msg.Length;
  MD.UpdateLoopNo:=0;
  Move(md5MagicNumber,MD.A,md5HashSize)
                                   {A,B,C,D mit Magic-Werten initialisieren}
end; {md5Init()}


procedure md5Update(MessageBufSize:longint);
  {MessageBufSize: Textlaenge, -> bei Dateien: Dateipuffer-Groesse (als Viel-
  faches von md5BlockSize!)}
type
  pChar=^Char;
var
  i:word;
begin {md5Update()}
  Inc(MD.UpdateLoopNo);
  i:=md5BlockSize;
  while (Msg.LengthRest>=md5BlockSize) and (i<=MessageBufSize) do begin
                                   {Vollstaendige MD5-Block lange}
                                   {Textabschnitte hashen}
    if (i=md5BlockSize) and (MD.UpdateLoopNo>1) then Msg.pRest:=Msg.pStart;
                                   {Lesen von Dateien: nach vollstaendigem Lesen}
                                   {des Dateipuffers zurueck auf Pufferanfang}
    Move(Msg.pRest^,MD.M,md5BlockSize);
                                   {Textabschnitt in MD5-Block kopieren}
    md5Transform;                  {MD5-Block hashen}
    Dec(Msg.LengthRest,md5BlockSize);
                                   {Textrestlaenge anpassen}
    Msg.pRest:=pChar(LongInt(Msg.pRest)+md5BlockSize);
                                   {Naechsten Textabschnitt vorwaehlen}
    Inc(i,md5BlockSize)
  end
end; {md5Update()}


procedure md5Finals(pHashValue:pointer);
  {pHashValue: Zeiger auf eine Variable von md5HashSize Laenge
  (128 Bit/16 Byte) zur Ruecklieferung des berechneten Hashwertes}

procedure AppPadBits(pM:pointer;AppStart,AppEnd:longint;var App1BitOk:boolean);
  {Padding Bits anfuegen}
type
  pChar=^Char;
var
  i:longint;
begin
  for i:=AppStart to AppEnd do begin
    if (i=AppStart) and not App1BitOk then begin
                                   {"1"-Bit setzen?}
      pChar(LongInt(pM)+i)^:=#$80; {"1"-Bit setzen ausfuehren}
      App1BitOk:=true
    end
    else pChar(LongInt(pM)+i)^:=#0 {"0"-Bits setzen ausfuehren}
  end
end;

var
  AppPad1BitOk:boolean;
begin {md5Finals()}
  Move(Msg.pRest^,MD.M,md5BlockSize);
                                   {Textabschnitt-Rest in MD5-Block kopieren}
  AppPad1BitOk:=false;
  AppPadBits(@MD.M,Msg.LengthRest,md5BlockSize-1,AppPad1BitOk);
                                   {Padding Bits anfuegen}
  if md5BlockSize-8<=Msg.LengthRest then begin
                                   {Kein Platz (8 Byte) mehr fuer}
                                   {Textlaengen-Zuweisung vorhanden?}
    md5Transform;                  {Zusaetzlichen MD5-Block hashen}
    AppPadBits(@MD.M,0,md5BlockSize-1,AppPad1BitOk);
                                   {Padding Bits fuer letzten Block anfuegen}
    Msg.LengthRest:=0
  end;
  MD.M[15]:=Msg.Length shl 3;      {Textlaenge (32 Bit) shl 3 (?) zuweisen,}
  MD.M[16]:=Msg.Length shr (32-3); {... durch shl 3 herausgefallene MSB-Anteile zuweisen}
  md5Transform;                    {letzten MD5-Block hashen}
  FillChar(MD.M,md5BlockSize,$00); {sensitive Daten im MD5-Block loeschen}
  Move(MD.A,pHashValue^,md5HashSize);
                                   {Hashwert von A,B,C,D nach pHashValue kopieren}
  FillChar(MD.A,md5HashSize,$00)   {sensitive Daten in A,B,C,D loeschen}
end; {md5Finals()}


function md5Valid:boolean;
  {Integritaetstest der MD5-Hashberechnung.
  Rueckgelieferter Funktionswert ist
  - true:  Hashwert-Berechnung entspricht Original-MD5-Algorithmus,
  - false: Quellcode veraendert oder TPU korrupt}
const
  TestStr:string[1]='a';
  TestStrHash='0cc175b9c0f1b6a831c399e269772661';
type
  Str2T=string[2];
var
  TestHashRet:array[1..md5HashSize] of byte;
  i:byte;

function HStr2Byte(s:Str2T):byte;
var n,i:byte;
begin
  n:=0;
  for i:=1 to 2 do begin
    n:=n shl 4;
    if s[i]<='9' then Inc(n,Ord(s[i]) and $0f) else Inc(n,(Ord(s[i]) and $0f)+9)
  end;
  HStr2Byte:=n
end;

begin
  md5Init(@TestStr[1],Length(TestStr));
  md5Update(Length(TestStr));
  md5Finals(@TestHashRet);
  md5Valid:=true;
  for i:=1 to md5HashSize do begin
    if TestHashRet[i]<>HStr2Byte(Copy(TestStrHash,i*2-1,2)) then md5Valid:=false
  end
end;


procedure md5Transform;
  {MD5: Berechnung jeweils eines 512-Bit-Blocks in MD.M und
  Addition der MD.A,B,C,D-Variablen mit dem Berechnungsergebnis.
  16Bit DOS-Implementierung 01.05/U.Gschwind}

function F(x,y,z:longint):longint;
  {Nichtlineare Funktion Nr. 1}
begin
  F:=(x and y) or (not x and z)
end;

function G(x,y,z:longint):longint;
  {Nichtlineare Funktion Nr. 2}
begin
  G:=(x and z) or (y and not z)
end;

function H(x,y,z:longint):longint;
  {Nichtlineare Funktion Nr. 3}
begin
  H:=x xor y xor z
end;

function I(x,y,z:longint):longint;
  {Nichtlineare Funktion Nr. 4}
begin
  I:=y xor (x or not z)
end;

function RotateLeft(x:longint;n:integer):longint;
  {Rotiert x n Bits nach links}
begin
  RotateLeft:=(x shl n) or (x shr (32-n))
end;

procedure FF(var a:longint;b,c,d,x:longint;s:integer;ac:longint);
  {Transformationsfunktion Nr. 1}
begin
  a:=RotateLeft(a+F(b,c,d)+x+ac,s)+b
end;

procedure GG(var a:longint;b,c,d,x:longint;s:integer;ac:longint);
  {Transformationsfunktion Nr. 2}
begin
  a:=RotateLeft(a+G(b,c,d)+x+ac,s)+b
end;

procedure HH(var a:longint;b,c,d,x:longint;s:integer;ac:longint);
  {Transformationsfunktion Nr. 3}
begin
  a:=RotateLeft(a+H(b,c,d)+x+ac,s)+b
end;

procedure II(var a:longint;b,c,d,x:longint;s:integer;ac:longint);
  {Transformationsfunktion Nr. 4}
begin
  a:=RotateLeft(a+I(b,c,d)+x+ac,s)+b
end;

const
  S11=7;S12=12;S13=17;S14=22;      {rotate left-Konstanten}
  S21=5;S22= 9;S23=14;S24=20;
  S31=4;S32=11;S33=16;S34=23;
  S41=6;S42=10;S43=15;S44=21;
var
  a,b,c,d:longint;

begin {md5Transform()}
  a:=MD.A;
  b:=MD.B;
  c:=MD.C;
  d:=MD.D;

    {Runde 1, Ergebnis der Prozeduraufrufe im ersten Parameter rueckgeliefert:}
  FF(a,b,c,d,MD.M[ 1],S11,$d76aa478);
  FF(d,a,b,c,MD.M[ 2],S12,$e8c7b756);
  FF(c,d,a,b,MD.M[ 3],S13,$242070db);
  FF(b,c,d,a,MD.M[ 4],S14,$c1bdceee);
  FF(a,b,c,d,MD.M[ 5],S11,$f57c0faf);
  FF(d,a,b,c,MD.M[ 6],S12,$4787c62a);
  FF(c,d,a,b,MD.M[ 7],S13,$a8304613);
  FF(b,c,d,a,MD.M[ 8],S14,$fd469501);
  FF(a,b,c,d,MD.M[ 9],S11,$698098d8);
  FF(d,a,b,c,MD.M[10],S12,$8b44f7af);
  FF(c,d,a,b,MD.M[11],S13,$ffff5bb1);
  FF(b,c,d,a,MD.M[12],S14,$895cd7be);
  FF(a,b,c,d,MD.M[13],S11,$6b901122);
  FF(d,a,b,c,MD.M[14],S12,$fd987193);
  FF(c,d,a,b,MD.M[15],S13,$a679438e);
  FF(b,c,d,a,MD.M[16],S14,$49b40821);

    {Runde 2:}
  GG(a,b,c,d,MD.M[ 2],S21,$f61e2562);
  GG(d,a,b,c,MD.M[ 7],S22,$c040b340);
  GG(c,d,a,b,MD.M[12],S23,$265e5a51);
  GG(b,c,d,a,MD.M[ 1],S24,$e9b6c7aa);
  GG(a,b,c,d,MD.M[ 6],S21,$d62f105d);
  GG(d,a,b,c,MD.M[11],S22, $2441453);
  GG(c,d,a,b,MD.M[16],S23,$d8a1e681);
  GG(b,c,d,a,MD.M[ 5],S24,$e7d3fbc8);
  GG(a,b,c,d,MD.M[10],S21,$21e1cde6);
  GG(d,a,b,c,MD.M[15],S22,$c33707d6);
  GG(c,d,a,b,MD.M[ 4],S23,$f4d50d87);
  GG(b,c,d,a,MD.M[ 9],S24,$455a14ed);
  GG(a,b,c,d,MD.M[14],S21,$a9e3e905);
  GG(d,a,b,c,MD.M[ 3],S22,$fcefa3f8);
  GG(c,d,a,b,MD.M[ 8],S23,$676f02d9);
  GG(b,c,d,a,MD.M[13],S24,$8d2a4c8a);

    {Runde 3:}
  HH(a,b,c,d,MD.M[ 6],S31,$fffa3942);
  HH(d,a,b,c,MD.M[ 9],S32,$8771f681);
  HH(c,d,a,b,MD.M[12],S33,$6d9d6122);
  HH(b,c,d,a,MD.M[15],S34,$fde5380c);
  HH(a,b,c,d,MD.M[ 2],S31,$a4beea44);
  HH(d,a,b,c,MD.M[ 5],S32,$4bdecfa9);
  HH(c,d,a,b,MD.M[ 8],S33,$f6bb4b60);
  HH(b,c,d,a,MD.M[11],S34,$bebfbc70);
  HH(a,b,c,d,MD.M[14],S31,$289b7ec6);
  HH(d,a,b,c,MD.M[ 1],S32,$eaa127fa);
  HH(c,d,a,b,MD.M[ 4],S33,$d4ef3085);
  HH(b,c,d,a,MD.M[ 7],S34, $4881d05);
  HH(a,b,c,d,MD.M[10],S31,$d9d4d039);
  HH(d,a,b,c,MD.M[13],S32,$e6db99e5);
  HH(c,d,a,b,MD.M[16],S33,$1fa27cf8);
  HH(b,c,d,a,MD.M[ 3],S34,$c4ac5665);

    {Runde 4:}
  II(a,b,c,d,MD.M[ 1],S41,$f4292244);
  II(d,a,b,c,MD.M[ 8],S42,$432aff97);
  II(c,d,a,b,MD.M[15],S43,$ab9423a7);
  II(b,c,d,a,MD.M[ 6],S44,$fc93a039);
  II(a,b,c,d,MD.M[13],S41,$655b59c3);
  II(d,a,b,c,MD.M[ 4],S42,$8f0ccc92);
  II(c,d,a,b,MD.M[11],S43,$ffeff47d);
  II(b,c,d,a,MD.M[ 2],S44,$85845dd1);
  II(a,b,c,d,MD.M[ 9],S41,$6fa87e4f);
  II(d,a,b,c,MD.M[16],S42,$fe2ce6e0);
  II(c,d,a,b,MD.M[ 7],S43,$a3014314);
  II(b,c,d,a,MD.M[14],S44,$4e0811a1);
  II(a,b,c,d,MD.M[ 5],S41,$f7537e82);
  II(d,a,b,c,MD.M[12],S42,$bd3af235);
  II(c,d,a,b,MD.M[ 3],S43,$2ad7d2bb);
  II(b,c,d,a,MD.M[10],S44,$eb86d391);

  MD.A:=MD.A+a;
  MD.B:=MD.B+b;
  MD.C:=MD.C+c;
  MD.D:=MD.D+d;

  md5TestShowMD(@MD.A,@MD.M,md5BlockSize,Msg.LengthRest)
                                   {Testmodus}
end; {md5Transform()}


procedure md5TestShowMD(pABCD,pMD5Block:pointer;md5BlockSize,Len:longint);
  {MD5: Nur zu Testzwecken, Anzeige von A,B,C,D und M.
  Bei Bedarf Kommentarklammern zwischen begin und end entfernen und Unit Crt einbinden.
  Wird die Prozedur zu keinem Zeitpunkt benoetigt, kann sie komplett entfernt werden.}
type
  Str2T=String[2];

function Byte2HStr(b:byte):Str2T;
const Hex:Array[0..15] of char='0123456789abcdef';
begin Byte2HStr:=Hex[(b shr 4)]+Hex[(b and $0f)] end;

type
  pChar=^Char;
var
  TextAttrLast,Y,i,j:byte;
  MChar:Char;
  AppLenReq:boolean;
begin
(*
  TextAttrLast:=TextAttr;
  Y:=WhereY;

    {A,B,C,D-Werte zeigen:}
  gotoXY(1,Y);
  write('>');
  for i:=0 to md5HashSize div 4-1 do begin
    if i>0 then write(' ');
    for j:=3 downto 0 do write(Byte2HStr(Ord(pChar(LongInt(pABCD)+(i*4)+j)^)))
  end;
  writeln;

    {MD5-Block-Werte zeigen:}
  AppLenReq:=md5BlockSize-8>Msg.LengthRest;
  gotoXY(1,Y+1);
  write('>');
  for i:=0 to md5BlockSize-1 do begin
    MChar:=pChar(LongInt(pMD5Block)+i)^;
    TextColor(00);TextBackground(02);
    if i>Len-1 then TextBackground(06);
    if i mod 8=0 then TextColor(15);
    if i<=Len-1 then begin
      gotoXY(i+2,Y+1);
      write(MChar)
    end
    else begin
      if (i<=md5BlockSize-1-8) or not AppLenReq then begin
        GotoXY(i+2,Y+1);
        if MChar=#$80 then write('1')
        else if MChar=#$00 then write('0')
        else write(MChar)
      end
      else begin
        TextBackground(03);
        if Odd(i) then TextBackground(01);
        write(Byte2HStr(Ord(MChar)))
      end
    end
  end;
  TextAttr:=TextAttrLast;
  write(' ',md5BlockSize,'/',Len);
  writeln
*)
end;

end.
